from setuptools import setup
setup(
    name='parmscheck',
    version='1.0.7',
    author='Lfr',
    author_email='1733535832@qq.com',
    description='Decorator for type checking arguments ',
    license = 'Apache',
    packages=['parmscheck'],
    url='https://github.com/violetDelia/parmscheck',
    install_requires=[
    ],
)